# encoding: utf-8

from datetime import date

__author__ = 'Nicolas Cadou'
__copyright__ = 'Copyright 2013-%d, Nicolas Cadou' % date.today().year
__credits__ = ['Nadeem Douba', 'Dan-Haim']

__license__ = 'New BSD'
__version__ = '0.1'
__maintainer__ = 'Nicolas Cadou'
__email__ = 'ncadou@cadou.ca'
__status__ = 'Development'
__url__ = 'http://pypi.python.org/pypi/onion-proctor'

__all__ = ['proxy', 'scripts', 'socket', 'tor']
